/* C support code for CLM */

/* TODO: these need arg range checks */

#include <config.h>

#if defined(SGI) || defined(LINUX) || defined(SUN) || defined(MAC_OSX)
  #include <sys/types.h>
  #include <sys/ipc.h>
  #include <sys/shm.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#if HAVE_FCNTL_H
  #include <fcntl.h>
#endif
#include <signal.h>
#if HAVE_LIMITS_H
  #include <limits.h>
#endif
#include <errno.h>

#if (defined(HAVE_LIBC_H) && (!defined(HAVE_UNISTD_H)))
  #include <libc.h>
#else
  #if (!(defined(_MSC_VER))) && (!(defined(MPW_C)))
    #include <unistd.h>
  #endif
#endif

#include "cmus.h"
#include "clm.h"

/* in MCL, fprintf causes the Mac to crash, so we go through mus_error2clm for all such calls */

#ifdef MCL_PPC
  void (*lisp_printf_callback)(char *);
  void (*lisp_error_callback)(void);
  void (*lisp_scheduler_callback)(void);
  void set_lisp_callbacks(void (*lp)(char *), void (*ep)(void), void(*sp)(void))
  {
    lisp_printf_callback = lp;
    lisp_error_callback = ep;
    lisp_scheduler_callback = sp;
  }
void clm_error(void) {(*lisp_error_callback)();}
#endif

static void mus_error2clm(int type, char *str)
{
#ifdef MCL_PPC 
  (*lisp_printf_callback)(str);
#else
  #ifdef ACL_LITE
    return;
  #else
    #if defined(WINDOZE) && defined(ACL4)
      aclprintf(str);
    #else
      fprintf(stdout, str); fflush(stdout);
    #endif
  #endif
#endif
}


/* ---------------- arrays ---------------- */

mus_sample_t setarray(mus_sample_t *arr, int i, mus_sample_t val) 
{
  arr[i] = val; 
  return(val);
}

mus_sample_t getarray(mus_sample_t *arr, int i) 
{
  return(arr[i]);
}

mus_sample_t incarray(mus_sample_t *arr, int i, mus_sample_t val) 
{
  arr[i] += val; 
  return(arr[i]);
}

mus_sample_t *makearray(int len) 
{
  mus_sample_t *ip; 
  ip = (mus_sample_t *)CALLOC(len, sizeof(mus_sample_t)); 
  return(ip);
}

void freearray(mus_sample_t *ip) {if (ip) FREE(ip);}

void cleararray1(int beg, int end, mus_sample_t *arr) 
{
  int i; 
  for (i=beg; i<=end; i++) arr[i] = MUS_SAMPLE_0;
}

void arrblt(int beg, int end, int newbeg, mus_sample_t *arr) 
{
  int i, j; 
  for (i=beg, j=newbeg; i>=end; i--, j--) arr[j]=arr[i];
}

mus_sample_t absmaxarr(int beg, int end, mus_sample_t *arr)
{
  mus_sample_t maxA, val;
  int i;
  maxA = MUS_SAMPLE_0;
  for (i=beg; i<=end; i++)
    {
      val = mus_sample_abs(arr[i]);
      if (val > maxA) maxA = val;
    }
  return(maxA);
}

/* this is to work around a bug in ACL 4.3 */
void acl43_set_arr(int *arr, int j, int a, int b, int c, int d);
void acl43_set_arr(int *arr, int j, int a, int b, int c, int d)
{
  arr[j] = a + (b<<8) + (c<<16) + (d<<24);
}
void acl43_get_arr(int *arr, int j, int *chr);
void acl43_get_arr(int *arr, int j, int *chr)
{
  chr[0] = ((arr[j]>>24) & 0xff);
  chr[1] = ((arr[j]>>16) & 0xff);
  chr[2] = ((arr[j]>>8) & 0xff);
  chr[3] = (arr[j] & 0xff);
}

void mus_write_float(int fd, float val) {write(fd, (char *)(&val), 4);}

#ifndef MACOS
sigfnc *clm_signal(int signo, sigfnc *fnc) {return(signal(signo, fnc));}
#endif

void print_io(int *io, int *datai)
{
  mus_error(MUS_NO_ERROR, "fd: %d, chans: %d, size: %d, beg: %d, end: %d, bufsiz: %d, data_start: %d, data_end: %d\n", 
	    io[CLM_IO_FD], io[CLM_IO_CHANS], io[CLM_IO_SIZE], io[CLM_IO_BEG], io[CLM_IO_END], io[CLM_IO_BUFSIZ], io[CLM_IO_DATA_START], io[CLM_IO_DATA_END]);
  mus_error(MUS_NO_ERROR, "dir: %d, loc: %d, hdr: %d, incr: %d, dats: %d, dats[0]: %d\n", 
	    io[CLM_IO_DIR], io[CLM_IO_LOC], io[CLM_IO_HDR_END], io[CLM_IO_INCR], io[CLM_IO_DATS], datai[io[CLM_IO_DATS]]);
}

static void clm_io_bufclr (int *io, int *datai, int beg)
{
  int i, k, end;
  mus_sample_t *j;
  if (io[CLM_IO_DATS+CLM_AREF_IBLOCK] == 0) 
    mus_error(MUS_BOGUS_FREE, "attempt to clear deallocated IO buffer");
  else
    {
      end=io[CLM_IO_BUFSIZ];
      for (k=0;k<io[CLM_IO_CHANS];k++)
	{
	  j=MUS_SAMPLE_ARRAY(datai[io[CLM_IO_DATS+CLM_AREF_IBLOCK]+k]);
	  if (j) {for (i=beg; i<end; i++) j[i]=MUS_SAMPLE_0;}
	}
    }
}

static int eof_unwarned = 1;

/* file_end, loc0 and loc should be longs */
void clm_file_reset(int loc0, int *io, int *datai)
{
  /* called when loc is outside the current in-core frame for the file pointed to by io */
  /* equivalent to the io.lisp function file-check + read-in */

  int file_end, bytes, loc;
  int bufend, filbytes;
#ifndef MACOS
  long lisp_call(int index);
#endif
  loc = loc0;
  if (io[CLM_IO_FD] == -1)
    {
      mus_error(MUS_FILE_CLOSED, "clm_file_reset of invalid file");
#ifndef MACOS
      print_io(io, datai);
      lisp_call(11);
#endif	  
      return;
    }
  if (io[CLM_IO_DIR] != CLM_IO_IN_FILE) 
    {
      if (io[CLM_IO_FD])
	{
	  if (io[CLM_IO_DATA_END] > io[CLM_IO_BUFSIZ])
	    mus_error(MUS_BUFFER_OVERFLOW, "data end indication in IO buffer is too big: %d > %d", io[CLM_IO_DATA_END], io[CLM_IO_BUFSIZ]);
	  else
	    {
	      if ((io[CLM_IO_DATA_START] < 0) || (io[CLM_IO_DATA_END] < 0)) 
		{
		  mus_error(MUS_BUFFER_UNDERFLOW, "file buffer index is negative! -- will try to fix it...");
		  if (io[CLM_IO_DATA_START] < 0) io[CLM_IO_DATA_START] = 0; else io[CLM_IO_DATA_END] = 0;
		}
	      mus_file_seek_frame(io[CLM_IO_FD], io[CLM_IO_BEG]+io[CLM_IO_DATA_START]);
	      mus_file_write(io[CLM_IO_FD], io[CLM_IO_DATA_START], io[CLM_IO_DATA_END], io[CLM_IO_CHANS], (mus_sample_t **)(datai+io[CLM_IO_DATS+CLM_AREF_IBLOCK]));
	      io[CLM_IO_DATA_START] = io[CLM_IO_DATA_END];
	      if (io[CLM_IO_SIZE] <= (io[CLM_IO_BEG] + io[CLM_IO_DATA_END])) io[CLM_IO_SIZE] = (io[CLM_IO_BEG] + 1 + io[CLM_IO_DATA_END]);
	      /* changed from < to <= 27-July-98 */
	    }
	}
    }
  if ((loc < io[CLM_IO_BEG]) && ((loc + (int)(.9*io[CLM_IO_BUFSIZ])) > io[CLM_IO_BEG]))
    {
      if ((loc + 10) > io[CLM_IO_BEG]) loc -= (int)(.75*io[CLM_IO_BUFSIZ]);
      if (loc < 0) loc = 0;
      if (io[CLM_IO_CHANS] == 1) loc = (2 * (int)(loc / 2));
    }
  file_end = io[CLM_IO_SIZE];
  bytes = file_end - loc;
  if (bytes > io[CLM_IO_BUFSIZ]) bytes=io[CLM_IO_BUFSIZ];
  if (bytes < 0)                   /* tried to access beyond current end of file */
    {
      if (io[CLM_IO_DIR] == CLM_IO_IN_FILE) 
	{
	  if (eof_unwarned)
	    {
	      mus_error(MUS_FILE_OVERFLOW, "attempt to read past EOF");
	      eof_unwarned = 0;
	    }
#ifndef MACOS
	  print_io(io, datai);
	  lisp_call(11);
#endif	  
	}
      else
	{
	  clm_io_bufclr(io, datai, 0);
	  bytes = io[CLM_IO_BUFSIZ];
	  io[CLM_IO_DATA_START] = 0;
	  io[CLM_IO_DATA_END] = 0;
	  lseek(io[CLM_IO_FD], 0, SEEK_END);
	  if (io[CLM_IO_CHANS] != 1)
	    {
	      mus_file_write_zeros(io[CLM_IO_FD], io[CLM_IO_CHANS]*(loc-file_end)); 
	      io[CLM_IO_BEG]=loc;
	    }
	  else
	    {
	      mus_file_write_zeros(io[CLM_IO_FD], loc-file_end); 
	      if ((loc%2)==0) io[CLM_IO_BEG]=loc; 
	      else io[CLM_IO_BEG]=loc-1;
	    }
	}
    }
  else /* bytes is positive or 0 */
    {
      /* fprintf(stderr, "seek %d (%d) ", io[CLM_IO_HDR_END]+(2*io[CLM_IO_CHANS]*loc), bytes); */
      mus_file_seek_frame(io[CLM_IO_FD], loc);
      io[CLM_IO_BEG] = loc;
      if (bytes > 0) 
	{
	  if (bytes > io[CLM_IO_BUFSIZ]) 
	    mus_error(MUS_BUFFER_OVERFLOW, "input request is too big: %d", bytes);
	  else
	    {
	      mus_file_read(io[CLM_IO_FD], 0, bytes-1, io[CLM_IO_CHANS], (mus_sample_t **)(datai+io[CLM_IO_DATS+CLM_AREF_IBLOCK]));
	    }
	}
      if (bytes < io[CLM_IO_BUFSIZ]) clm_io_bufclr(io, datai, bytes);
      io[CLM_IO_DATA_START] = 0;
      io[CLM_IO_DATA_END] = 0;
      if (io[CLM_IO_DIR] == CLM_IO_IN_FILE)
	{
	  bufend = io[CLM_IO_BUFSIZ]-1;
	  filbytes = file_end-io[CLM_IO_BEG]-1;
	  if (filbytes < bufend) bufend = filbytes;
	  if (bufend > 0) io[CLM_IO_DATA_END] = bufend;
	}
    }
  io[CLM_IO_END] = io[CLM_IO_BEG]+io[CLM_IO_BUFSIZ]-1;
}

/* loc should be long, as should io[beg], io[end] */
void clm_full_file_reset(int loc, int *io, int *datai)
{
  /* update data_end before calling clm_file_reset -- we are about to write at loc, so loc-1 is last output */
  /* dangerous assumption!? */
  int lim;
  lim = ((loc - io[CLM_IO_BEG]) - 1);
  if (lim >= io[CLM_IO_BUFSIZ]) lim = (io[CLM_IO_BUFSIZ]-1);
  if (lim > io[CLM_IO_DATA_END]) io[CLM_IO_DATA_END] = lim;
  clm_file_reset(loc, io, datai);
}


double clm_oscil_bank(double *amps, double *oscils, double *inputs, int size)
{
  int i, j;
  double sum = 0.0;
  for (i=0, j=0; i<size; i++, j+=2) 
    {
      sum += (amps[i] * sin(oscils[j+1]));
      oscils[j+1] += (oscils[j] + inputs[i]);
      if ((oscils[j+1] > 100.0) || (oscils[j+1]<-100.0)) oscils[j+1] = fmod(oscils[j+1], TWO_PI);
    }
  return(sum);
}


/* rand taken from the ANSI C standard (essentially the same as the Cmix form used earlier) */
static unsigned long randx = 1;
#define INVERSE_MAX_RAND  0.0000610351563
#define INVERSE_MAX_RAND2 0.000030517579

void clm_srand(int val) {randx = val;}

double clm_frandom(double amp) /* -amp to amp as double */
{
  int val;
  randx=randx*1103515245 + 12345;
  val=(unsigned int)(randx >> 16) & 32767;
  return(amp * (((double)val)*INVERSE_MAX_RAND - 1.0));
}

double clm_random(double amp) /* 0..amp as double */
{
  int val;
  randx=randx*1103515245 + 12345;
  val=(unsigned int)(randx >> 16) & 32767;
  return(amp * (((double)val)*INVERSE_MAX_RAND2));
}

double clm_random_any(double amp, double *distribution, int distribution_size) /* -amp to amp through distribution */
{
  return(amp * clm_table_interp(distribution, clm_random((double)distribution_size), distribution_size));
}

/* on 32-bit long machine, seed = 1664525L*seed + 1013904223L; */
/* (Knuth -- probably as good as above clm_frandom */


/* TODO: do these substitutions and libclm.def the clm.h stuff */
double clm_table_interp (double *fn, double xx, int len)
{
  return(mus_array_interp(fn, xx, len));
}

void clm_fft (double *rl, double *im, int n, int is)
{
  mus_fft(rl, im, n, is);
}

void clm_convolve (double* rl1, double* rl2, int n)
{
  mus_convolution(rl1, rl2, n);
}

void clm_spectrum(double *rdat, double *idat, double *window, int n, int type)
{
  mus_spectrum(rdat, idat, window, n, type);
}


/* parallel processing scheduler for "p" instruments
 *
 * Each such instrument uses C macros for the variable references so that they remain pointers into the data arrays.
 * The arrays are saved when the instrument's current note starts, along with an id (a back-pointer), and the 
 *   instrument's loop-function entry point and its current sample.  The scheduler is then invoked, 
 *   and once all waiting p-instruments have run up to (not including) the current sample, control returns to Lisp.
 * Upon with-sound end (or clm-reset), the waiting processes are run to completion.
 * Each loop of the scheduler sets up the data array pointers and invokes the loop-function.
 * Upon loop completion, the instrument calls its end-function, removes itself from the scheduler pool, and 
 *   frees its data arrays.  If any local files are open, their indices are added to pool_open_index for later closing.
 *   I believe it is safe to leave these io blocks untended because each separate pooled process will have called
 *   open-input, creating its own buffers and so on.
 * The saved arrays (+ the sample counter) constitute the process state.
 *
 * If the scheduler finds itself without anything to run, it resets its "current sample" to the waiting one's and returns.
 * If no one is waiting, it returns.
 * Only a p-instrument and clu-reset call the scheduler -- it is invisible to with-sound and friends.
 *
 * The other complication involves Ctrl-C interrupts -- if one is seen the scheduler tells each running
 *   instrument to free its arrays and clean up, then returns.  No further scheduling is allowed.
 * 
 * changed 11-July-95 for run-time allocation of arrays (SGI limitations)
 * changed 12-Feb-96 for direct-to-DAC support -- main change involves reverb input.  Since we can't currently ensure
 *   that the reverb runs last (and thereby picks up the accumulated reverb input for this sample), we save that
 *   value in the 2nd reverb pool_outn slot.  The previn generator (rt.lisp) then gets and clears the value.  The
 *   rest of CLM does not know whether the output is to a file or to the DAC -- the decision is made only at
 *   checked_write in io.c based on the file descriptor value -- DAC_CHANNEL means send output buffer to DAC.
 * changed 14-July-97 for realtime controls and whatnot.
 */

#define POOL_SIZE 256              /* max number of simultaneously active parallel instruments                           */
#define CLM_IO_CHANNELS 64         /* must be at least == *available-io-channels* in defaults.lisp                       */
#define CLM_SOUND_CHANNELS 66      /* max chans per sound + 2 for reverb                                                 */
#define CLM_RECORD_CHANNELS 4
#ifndef CLM_REVERB_CHANNEL
  #define CLM_REVERB_CHANNEL 64
#endif

#define max_ins pool_globals[0]           /* top active instrument                                                       */
#define current_sample pool_globals[1]    /* just a convenience (scheduler's explicit notion of where it is)             */
#define pool_active pool_globals[2]
#define pool_max_sample pool_globals[3]   /* for final wait -- merely a convenience                                      */

static int pool_got_sigint = 0;    /* catch C-C and exit after immediate cleanup                                         */
static void pool_sig_err(int sig) {pool_got_sigint = sig;}

/* typedef int (*intpfun)(); */
typedef int (*intpfun)(int id, int samp, int end, int *dati, double *datr, double *outn, double *recn);

typedef int *pdatai;
typedef double *pdatar;

static int *pool_id;               /* back pointer for caller's benefit, 0==unused                                       */
static intpfun *pool_loop;         /* entry points of p-instrument loop functions                                        */
static pdatai *pool_datai;         /* pointers to int data arrays (malloc'd and later free'd by the p-instrument itself) */
static pdatar *pool_datar;         /* ditto for double data arrays                                                       */
static int *pool_sample;           /* waiting p-instrument instantiation's sample counter value                          */
static int *pool_end;
static double *pool_outn;          /* outa..d, reva for standard output collection                                       */
static double *pool_recn;          /* record channels                                                                    */
static int *pool_globals = NULL;   /* state variables for pool processes                                                 */

static void create_pool (void)
{
  pool_id = (int *)CALLOC(POOL_SIZE, sizeof(int));
  pool_sample = (int *)CALLOC(POOL_SIZE, sizeof(int));
  pool_end = (int *)CALLOC(POOL_SIZE, sizeof(int));
  pool_outn = (double *)CALLOC(CLM_SOUND_CHANNELS, sizeof(double));
  pool_recn = (double *)CALLOC(CLM_RECORD_CHANNELS, sizeof(double));
  pool_loop = (intpfun *)CALLOC(POOL_SIZE, sizeof(intpfun));
  pool_datai = (pdatai *)CALLOC(POOL_SIZE, sizeof(pdatai));
  pool_datar = (pdatar *)CALLOC(POOL_SIZE, sizeof(pdatar));
  pool_globals = (int *)CALLOC(4, sizeof(int));
  if ((pool_id == NULL) || (pool_end == NULL) || (pool_sample == NULL) || (pool_outn == NULL) ||
      (pool_loop == NULL) || (pool_recn == NULL) || (pool_datai == NULL) || (pool_datar == NULL) || (pool_globals == NULL))
    mus_error(MUS_MEMORY_ALLOCATION_FAILED, "pool array allocation trouble");
  max_ins=-1; current_sample=0; pool_active=0; pool_max_sample=0;
}

/* If all these are exported, p-instruments can examine other's internal variables, but that's for the future */
/* To implement this: return the id from initialize_ins (int decl passed back to lisp as result of init call) */
/*   with (perhaps) a hash table keyed to id giving ins name and its args. Add procs to read (and if desired, */
/*   set) the current pool array values given id, perhaps mapping functions over ins name or whatever.        */
/*   Would also need a way to distinguish the current process (i.e. run loop needs named access to its id)    */

void clm_initialize_pool (void)   /* make sure it's in a clean state, but do as little as possible */
{                             /* (this is called at clm initialization) */
  max_ins = -1;
  current_sample = 0;
  pool_max_sample = 0;
  pool_got_sigint = 0;
}

static int new_pool_id (void)        /* get free location in pool, -1=failure. */
{
  int i;
  if (pool_active == 0) 
    {
      for (i=0; i<POOL_SIZE; i++) pool_id[i] = 0; 
      pool_active = 1;
    }
  for (i=0; i<POOL_SIZE; i++) if (pool_id[i] == 0) {pool_id[i] = 1; return (i);}
  return(-1);
}

static int run_ins (int id)
{
  int ins_done;
  ins_done = (*pool_loop[id])(id, pool_sample[id], pool_end[id], pool_datai[id], pool_datar[id], pool_outn, pool_recn);
  pool_sample[id]++;
  return(ins_done);
}

void clm_finalize_ins (int id)
{
  pool_id[id] = 0;           /* mark pool[id] as free (p-instrument call is finished) */
  FREE(pool_datai[id]);
  FREE(pool_datar[id]);
  if (id == max_ins)
    {
      max_ins--;
      while ((max_ins >= 0) && (pool_id[max_ins] == 0)) max_ins--;
    }
}

static double read_adc(void);
static int adc_chans(void);

static int drive_on = 0;

static void wait_until (int sample, int *outn, int *rev, int *datai) 
  /* run all active p-instruments up to but not including the current sample */
{                            
  int i, oi, chans, rec_chans;
#ifndef MACOS
  void (*old_SIGINT)(int);
  long lisp_call(int index);
#endif
  chans = outn[CLM_IO_CHANS];
  rec_chans = adc_chans();
  if (pool_got_sigint == 0)
    {
#ifndef MACOS
      old_SIGINT = clm_signal(SIGINT, pool_sig_err); /* trap SIGINT */
#endif
      if ((outn) && (current_sample < outn[CLM_IO_BEG])) clm_file_reset(current_sample, outn, datai);
      if ((rev) && (current_sample < rev[CLM_IO_BEG])) clm_file_reset(current_sample, rev, datai);
      while (current_sample < sample)
	{
  	  if ((drive_on) || (max_ins > -1))           /* is there anyone in the pool yet? */
	    {
	      if ((outn) && (current_sample > outn[CLM_IO_END])) clm_full_file_reset(current_sample, outn, datai);
	      if ((rev) && (current_sample > rev[CLM_IO_END])) clm_full_file_reset(current_sample, rev, datai);
	      for (i=0; i<chans; i++) pool_outn[i]=0.0;
	      for (i=0; i<rec_chans; i++) pool_recn[i] = read_adc();
	      pool_outn[CLM_REVERB_CHANNEL]=0.0;
	      for (i=0; i<=max_ins; i++)
		{                     /* run any active p-ins, end any that are up to their final sample */
		  if (pool_id[i] != 0) {if (run_ins(i)) clm_finalize_ins(i);}
		}
	      oi=current_sample-outn[CLM_IO_BEG];
	      for (i=0; i<chans; i++)
		{
		  (((mus_sample_t *)(datai[outn[CLM_IO_DATS+CLM_AREF_IBLOCK]+i]))[oi]) += (mus_sample_t)(pool_outn[i]);
		}
	      if (outn[CLM_IO_DATA_END] < oi) outn[CLM_IO_DATA_END] = oi;
	      if (rev)
		{
		  oi=current_sample-rev[CLM_IO_BEG];
		  pool_outn[CLM_REVERB_CHANNEL+1] = pool_outn[CLM_REVERB_CHANNEL];
		  (((mus_sample_t *)(datai[rev[CLM_IO_DATS+CLM_AREF_IBLOCK]]))[oi]) += (mus_sample_t)(pool_outn[CLM_REVERB_CHANNEL]);
		  if (rev[CLM_IO_DATA_END] < oi) rev[CLM_IO_DATA_END] = oi;
		}
	      current_sample++;
	      if (pool_got_sigint != 0)
		{    /* Ctrl-C so clean out any active instruments and warn lisp */
		  for (i=0; i<=max_ins; i++)
		    { 
		      if (pool_id[i] != 0) clm_finalize_ins(i);
		    }
		  current_sample = sample;
#ifndef MACOS
		  clm_signal(SIGINT, old_SIGINT);
		  lisp_call(13);
		  old_SIGINT = clm_signal(SIGINT, pool_sig_err);
#endif
		}
	    } 
	  else                    /* no one is able to run, so we might as well jump to sample and wait */
	    {
	      current_sample = sample;
	    }
	}
#ifndef MACOS
      clm_signal(SIGINT, old_SIGINT);
#endif
    }
  else current_sample = sample;
}

int clm_driving(void) {return(drive_on);}
void clm_set_driving(int on) {drive_on = on;}

/* int clm_pool_active(void){return((pool_globals) && (pool_active));} */
 
int clm_temporize_pool (int sample, int out_adr, int rev_adr, int *datai)
{
  int *reva;
  if (sample > current_sample)
    {
      if (!rev_adr)
	reva = NULL;
      else reva = (int *)(datai+rev_adr);
      wait_until(sample, (int *)(datai+out_adr), reva, datai);
      
    }
  return(current_sample);
}

int clm_finalize_pool (int out_adr, int rev_adr, int *datai)
{
  int *reva;
  int io_changed = 0;
  if ((pool_globals) && (pool_active))
    {
      if (!rev_adr)
	reva = NULL;
      else reva = (int *)(datai+rev_adr);
      wait_until(pool_max_sample+1, (int *)(datai+out_adr), reva, datai);

      clm_file_reset(0, (int *)(datai+out_adr), datai);
      if (reva) clm_file_reset(0, reva, datai);
      io_changed = 1;
    }

  pool_active = 0;
  pool_max_sample = 0;
  pool_got_sigint = 0;
  eof_unwarned = 1;
  return(io_changed);
}

void clm_reset_pool (void)
{
  /* in case of premature exit (via error reset in with-sound, for example), clean up the pool */
  int i;
  if ((pool_globals) && (pool_active) && (pool_got_sigint == 0))
    {
      for (i=0; i<=max_ins; i++)
	{ 
	  if (pool_id[i] != 0) clm_finalize_ins(i);
	}
    }
  max_ins = -1;
  current_sample = 0;
  pool_active = 0;
  pool_max_sample = 0;
  pool_got_sigint = 0;
}

void clm_initialize_ins (int beg_sample, int end_sample, int out_adr, int rev_adr, 
		     int *datai, int leni, double *datar, int lenr, 
		     int (*loop_function)(int id, int samp, int end, int *dati, double *datr, double *outn, double *recn))
{
  int id, i;
  int *di, *rev, *outn;
  double *dr;
  if (pool_got_sigint == 0)
    {
      if (!rev_adr) 
	rev = NULL;
      else rev = (int *)(datai+rev_adr);
      if (!out_adr)
	outn = NULL;
      else outn = (int *)(datai+out_adr);
      if (beg_sample < current_sample) /* added 9-Mar-00 */
	{
	  end_sample += (current_sample - beg_sample);
	  beg_sample = current_sample;
	}
      if ((outn) || (rev))
	wait_until(beg_sample, outn, rev, datai);
                /* run anything in the pool up to (not including) beg_sample */
      id = new_pool_id();
      if (id != -1)
	{
	  di = (int *)CALLOC(leni, sizeof(int));
	  if (di == NULL) {mus_error(MUS_MEMORY_ALLOCATION_FAILED, "failed to allocate p-ins ints!"); pool_id[id] = 0; return;}
	  for (i=0; i<leni; i++) di[i] = datai[i];
	  dr = (double *)CALLOC(lenr, sizeof(double));
	  if (dr == NULL) {mus_error(MUS_MEMORY_ALLOCATION_FAILED, "failed to allocate p-ins doubles!"); FREE(di); pool_id[id] = 0; return;}
	  for (i=0; i<lenr; i++) dr[i] = datar[i];
	  if (pool_max_sample < end_sample) pool_max_sample = end_sample;
	  pool_loop[id] = loop_function;
	  pool_sample[id] = beg_sample;
          pool_end[id] = end_sample;
	  if (id > max_ins) max_ins = id;
	  pool_datai[id] = di;
	  pool_datar[id] = dr;
	}
      else mus_error(MUS_BUFFER_OVERFLOW, "p-ins pool is full!");
    }
}


#if defined(SGI) || defined(LINUX) || defined(SUN) || defined(MAC_OSX)

/* real time shared memory handlers */

static double *shm_base = NULL;
static int shm_fd = -1;
static int shm_size_in_bytes = 0;

double clm_control (int n)
{
  if (shm_base) return(shm_base[n]);
  return(0.0);
}

double clm_set_control (int type, int n, double val)
{
  if (shm_base)
    {
      if (type == 0) shm_base[n] = val;
      else if (type == 1) shm_base[n] += val;
      else shm_base[n] -= val;
    }
  return(val);
}

int clm_control_size(void) {return(shm_size_in_bytes);}

int clm_open_controls (int size)
{
  if (!shm_base)
    {
      shm_fd = shmget(CLM_SHM_KEY, size*sizeof(double), IPC_CREAT | IPC_EXCL | SHM_R | SHM_W);
                                              /* controller calls: fd = shmget(CLM_SHM_KEY, clm_control_size(), SHM_R | SHM_W); */
      if (shm_fd == -1) 
	{
	  if (errno == EEXIST)
	    shm_fd = shmget(CLM_SHM_KEY, size*sizeof(double), SHM_R | SHM_W);
	  if (shm_fd == -1)
	    {
	      perror("clm get shared memory");
	      return(-1);
	    }
	}
      shm_base = (double *)shmat(shm_fd, 0, 0);  /* controller also makes this call */
      if (((void *)shm_base) == ((void *)-1)) 
	{
	  shmctl(shm_fd, IPC_RMID, 0);
	  shm_base = NULL;
	  perror("clm attach shared memory");
	  return(-1);
	}
      shm_size_in_bytes = size*sizeof(double);
    }
  return(0);
}

void clm_close_controls (void)
{
  int err;
  if (shm_base)
    {
#if defined(LINUX) || defined(SUN)
      shmdt((char *)shm_base);
#else
      shmdt(shm_base);                       /* same in controller */
#endif
      err = shmctl(shm_fd, IPC_RMID, 0);
      shm_base = NULL;
      if (err == -1) perror("clm free shared memory"); /* not called in controller */
    }
}

int clm_attach_controls (int size)
{
  if (!shm_base)
    {
      shm_fd = shmget(CLM_SHM_KEY, size*sizeof(double), SHM_R | SHM_W);
      if (shm_fd == -1)
	{
	  perror("clm get shared memory");
	  return(-1);
	}
      shm_base = (double *)shmat(shm_fd, 0, 0);
      shm_size_in_bytes = size*sizeof(double);
    }
  return(0);
}

void clm_detach_controls (void)
{
  if (shm_base)
    {
#if defined(LINUX) || defined(SUN)
      shmdt((char *)shm_base);
#else
      shmdt(shm_base);
#endif
    }
}

#endif

#if ((defined(ACL4) || defined(CMU)) && (!(defined(WINDOZE))))

void clm_kill(int pid) 
{
  int err; 
  err = kill(pid, SIGKILL); 
  if (err == -1) 
    fprintf(stderr, "clm_kill: %s", strerror(errno));
}

#define MAXSIZE 256
static char strftime_buf[MAXSIZE];

char *clm_strftime (char *format)
{
#if HAVE_STRFTIME
  time_t ts;
  time(&ts);
  strftime(strftime_buf, MAXSIZE, format, localtime(&ts));
  return(strftime_buf);
#else
  return(NULL);
#endif
}
  
char *clm_errno_str(void)
{
  strcpy(strftime_buf, strerror(errno));
  return(strftime_buf);
}
#endif

/* output */
static int audio_fd = -1;

void close_dac (void) 
{
  if (audio_fd != -1) mus_audio_close(audio_fd);
  audio_fd = -1;
}

int open_dac (int srate, int channels, int size, int rchans)
{
  if (audio_fd != -1) close_dac();
  if (rchans == 0)
    audio_fd = mus_audio_open_output(MUS_AUDIO_DEFAULT, srate, channels, MUS_COMPATIBLE_FORMAT, size);
  else audio_fd = mus_audio_open_output(MUS_AUDIO_DUPLEX_DEFAULT, srate, channels, MUS_COMPATIBLE_FORMAT, size);
  return(audio_fd);
}

/* input */
static int record_fd = -1;
static int record_i = 0;
static int record_size = 0;
static char *record_buf = NULL;
static mus_sample_t *record_samps = NULL;
static int record_chans = 0;

static void reset_dacs(void) 
{
  record_size = 0; 
  record_chans = 0;
  audio_fd = -1;
  record_fd = -1;
  record_i = 0;
  record_buf = NULL;
  record_samps = NULL;
}

static int adc_chans(void) {return(record_chans);}

int open_adc (int srate, int channels, int line, int size)
{
  int dev;
  float vals[1];
  record_chans = channels;
  if (record_chans > 0)
    {
      if (record_buf) FREE(record_buf);
      record_buf = (char *)CALLOC(size, 4);
      if (record_samps) FREE(record_samps);
      record_samps = (mus_sample_t *)CALLOC(size, sizeof(mus_sample_t));
      record_size = size;
      if (line == 1) dev = MUS_AUDIO_LINE_IN; else {if (line == 2) dev = MUS_AUDIO_DIGITAL_IN; else dev = MUS_AUDIO_MICROPHONE;}
      mus_audio_mixer_read(dev, MUS_AUDIO_AMP, 0, vals);
      if (vals[0] == 0.0)
	{
	  vals[0]=1.0;
	  mus_audio_mixer_write(dev, MUS_AUDIO_AMP, 0, vals);
	}
#ifdef LINUX
      record_fd = mus_audio_open_input(MUS_AUDIO_DUPLEX_DEFAULT, srate, channels, MUS_COMPATIBLE_FORMAT, 0);
      /* mus_audio_mixer_write(MUS_AUDIO_DUPLEX_DEFAULT, MUS_AUDIO_PORT, dev, NULL); */
#else
      record_fd = mus_audio_open_input(dev, srate, channels, MUS_COMPATIBLE_FORMAT, 0);
#endif
    }
  record_i = record_size; /* triggers initial read */
  return(record_fd);
}

void close_adc (void)
{
  if (record_chans > 0)
    {
      record_chans = 0;
      mus_audio_close(record_fd);
      record_fd = -1;
    }
}

static double read_adc(void)
{
  /* called once per channel per sample */
  if (record_i == record_size)
    {
      record_i = 0;
      mus_audio_read(record_fd, record_buf, record_size*mus_bytes_per_sample(MUS_COMPATIBLE_FORMAT));
      mus_file_read_buffer(MUS_COMPATIBLE_FORMAT, 0, 1, record_size, &record_samps, record_buf);
      /* 0=beg, 1=chans */
    }
  return(MUS_SAMPLE_TO_DOUBLE(record_samps[record_i++]));
}

void initialize_cmus (void)
{
  create_pool();
  mus_error_set_handler(mus_error2clm);
  record_samps = NULL;
}

void reset_cmus(void)
{
  eof_unwarned = 1;
  reset_dacs();
  record_samps = NULL;
}

#define MIX_BUFFER_SIZE 8192

/* TODO: can clm_simple_mix be replaced by mus_mix? */

int clm_simple_mix(char *outfile, char *infile, int out_start, int out_frames, int in_start, double scaler, double prescaler)
{
  int i, j, k, ofd, ifd, in_chans, out_chans, curoutframes, datsize, min_chans;
  mus_sample_t **obufs, **ibufs;
  out_chans = mus_sound_chans(outfile);
  in_chans = mus_sound_chans(infile);
  if (out_chans < in_chans) min_chans = out_chans; else min_chans = in_chans;
  obufs = (mus_sample_t **)CALLOC(out_chans, sizeof(mus_sample_t *));
  for (i=0; i<out_chans; i++) obufs[i] = (mus_sample_t *)CALLOC(MIX_BUFFER_SIZE, sizeof(mus_sample_t));
  ibufs = (mus_sample_t **)CALLOC(in_chans, sizeof(mus_sample_t *));
  for (i=0; i<in_chans; i++) ibufs[i] = (mus_sample_t *)CALLOC(MIX_BUFFER_SIZE, sizeof(mus_sample_t));
  ifd = mus_sound_open_input(infile);
  if (prescaler > 0.0) mus_file_set_prescaler(ifd, (float)prescaler);
  mus_sound_seek_frame(ifd, in_start);
  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, in_chans, ibufs);
  ofd = mus_sound_reopen_output(outfile, out_chans, mus_sound_data_format(outfile), mus_sound_header_type(outfile), mus_sound_data_location(outfile));
  datsize = mus_bytes_per_sample(mus_sound_data_format(outfile));
  curoutframes = mus_sound_frames(outfile);
  if (curoutframes < out_start)
    {
      lseek(ofd, 0, SEEK_END);
      mus_file_write_zeros(ofd, (out_start - curoutframes)*out_chans);
    }
  mus_sound_seek_frame(ofd, out_start);
  mus_sound_read(ofd, 0, MIX_BUFFER_SIZE-1, out_chans, obufs);
  mus_sound_seek_frame(ofd, out_start);
  if (scaler == 1.0)
    {
      if (in_chans == 1)
	{
	  for (k=0, j=0;k<out_frames;k++, j++)
	    {
	      if (j == MIX_BUFFER_SIZE)
		{
		  mus_sound_write(ofd, 0, j-1, out_chans, obufs);
		  j = 0;
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ofd, 0, MIX_BUFFER_SIZE-1, out_chans, obufs);
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, in_chans, ibufs);
		}
	      obufs[0][j] += ibufs[0][j];
	    }
	}
      else
	{
	  for (k=0, j=0;k<out_frames;k++, j++)
	    {
	      if (j == MIX_BUFFER_SIZE)
		{
		  mus_sound_write(ofd, 0, j-1, out_chans, obufs);
		  j = 0;
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ofd, 0, MIX_BUFFER_SIZE-1, out_chans, obufs);
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, in_chans, ibufs);
		}
	      for (i=0; i<min_chans; i++) obufs[i][j] += ibufs[i][j];
	    }
	}
    }
  else
    {
      if (in_chans == 1)
	{
	  for (k=0, j=0;k<out_frames;k++, j++)
	    {
	      if (j == MIX_BUFFER_SIZE)
		{
		  mus_sound_write(ofd, 0, j-1, out_chans, obufs);
		  j = 0;
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ofd, 0, MIX_BUFFER_SIZE-1, out_chans, obufs);
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, in_chans, ibufs);
		}
	      obufs[0][j] += (mus_sample_t)(scaler * ibufs[0][j]);
	    }
	}
      else
	{
	  for (k=0, j=0;k<out_frames;k++, j++)
	    {
	      if (j == MIX_BUFFER_SIZE)
		{
		  mus_sound_write(ofd, 0, j-1, out_chans, obufs);
		  j = 0;
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ofd, 0, MIX_BUFFER_SIZE-1, out_chans, obufs);
		  mus_sound_seek_frame(ofd, out_start+k);
		  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, in_chans, ibufs);
		}
	      for (i=0; i<min_chans; i++) obufs[i][j] += (mus_sample_t)(ibufs[i][j] * scaler);
	    }
	}
    }
  if (j > 0) mus_sound_write(ofd, 0, j-1, out_chans, obufs);
  if (curoutframes < (out_frames+out_start)) curoutframes = out_frames + out_start;
  mus_sound_close_output(ofd, curoutframes * out_chans * datsize);
  mus_sound_close_input(ifd);
  for (i=0; i<in_chans; i++) FREE(ibufs[i]);
  FREE(ibufs);
  for (i=0; i<out_chans; i++) FREE(obufs[i]);
  FREE(obufs);
  return(curoutframes);
}

int clm_scale_file(char *outfile, char *infile, double scaler, int out_format, int out_header)
{
  int i, j, k, ofd, ifd, chans, datsize, frames;
  mus_sample_t **obufs;
  chans = mus_sound_chans(infile);
  obufs = (mus_sample_t **)CALLOC(chans, sizeof(mus_sample_t *));
  for (i=0; i<chans; i++) obufs[i] = (mus_sample_t *)CALLOC(MIX_BUFFER_SIZE, sizeof(mus_sample_t));
  ifd = mus_sound_open_input(infile);
  mus_sound_seek_frame(ifd, 0);
  mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, chans, obufs);
  ofd = mus_sound_open_output(outfile, mus_sound_srate(infile), chans, out_format, out_header, NULL);
  datsize = mus_bytes_per_sample(out_format);
  frames = mus_sound_frames(infile);
  if (chans == 1)
    {
      for (k=0, j=0;k<frames;k++, j++)
	{
	  if (j == MIX_BUFFER_SIZE)
	    {
	      mus_sound_write(ofd, 0, j-1, chans, obufs);
	      j = 0;
	      mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, chans, obufs);
	    }
	  obufs[0][j] = (mus_sample_t)(scaler * obufs[0][j]);
	}
    }
  else
    {
      for (k=0, j=0;k<frames;k++, j++)
	{
	  if (j == MIX_BUFFER_SIZE)
	    {
	      mus_sound_write(ofd, 0, j-1, chans, obufs);
	      j = 0;
	      mus_sound_read(ifd, 0, MIX_BUFFER_SIZE-1, chans, obufs);
	    }
	  for (i=0; i<chans; i++) obufs[i][j] = (mus_sample_t)(obufs[i][j] * scaler);
	}
    }
  if (j > 0) mus_sound_write(ofd, 0, j-1, chans, obufs);
  mus_sound_close_output(ofd, frames * chans * datsize);
  mus_sound_close_input(ifd);
  for (i=0; i<chans; i++) FREE(obufs[i]);
  FREE(obufs);
  return(frames);
}

int clm_file2array(char *filename, int chan, int start, int samples, double *array)
{
  mus_sample_t *idata;
  int i, len;
  idata = (mus_sample_t *)CALLOC(samples, sizeof(mus_sample_t));
  len = mus_file_to_array(filename, chan, start, samples, idata);
  if (len != -1) for (i=0; i<samples; i++) array[i] = MUS_SAMPLE_TO_DOUBLE(idata[i]);
  FREE(idata);
  return(len);
}

void clm_array2file(char *filename, double *ddata, int len, int srate, int channels)
{
  mus_sample_t *idata;
  int i;
  idata = (mus_sample_t *)CALLOC(len, sizeof(mus_sample_t));
  for (i=0; i<len; i++) idata[i] = MUS_DOUBLE_TO_SAMPLE(ddata[i]);
  mus_array_to_file(filename, idata, len, srate, channels);
  FREE(idata);
}

int clm_sound_loop_info(char *name, int *vals)
{
  /* need a wrapper since ACL can't handle returning int* */
  int i;
  int *rtn;
  rtn = mus_sound_loop_info(name);
  if (rtn)
    {
      for (i=0; i<6; i++) vals[i] = rtn[i];
      return(1);
    }
  return(0);
}


/* sndplay as function call, added 14-Jul-00, 
 *                           fixed up mus_file_read trailing chunk business 26-Jan-01
 *                           added device arg 8-Feb-01
 */

#if MAC_OSX
  #define BUFFER_SIZE 256
#else
  #define BUFFER_SIZE 4096
#endif

int sl_dac(char *name, int output_device)
{
  int fd, afd, i, j, n, k, chans, srate, frames, outbytes, curframes;
  mus_sample_t **bufs;
  short *obuf;
  mus_sound_initialize();

  afd = -1;
  fd = mus_sound_open_input(name);
  if (fd != -1)
    {
      chans = mus_sound_chans(name);
      srate = mus_sound_srate(name);
      frames = mus_sound_frames(name);
      outbytes = BUFFER_SIZE * chans * 2;
      bufs = (mus_sample_t **)CALLOC(chans, sizeof(mus_sample_t *));
      for (i=0; i<chans; i++) bufs[i] = (mus_sample_t *)CALLOC(BUFFER_SIZE, sizeof(mus_sample_t));
      obuf = (short *)CALLOC(BUFFER_SIZE * chans, sizeof(short));
      /* assume for lafs that our DAC wants 16-bit integers */
      for (i=0; i<frames; i+=BUFFER_SIZE)
	{
	  if ((i+BUFFER_SIZE) <= frames)
	    curframes = BUFFER_SIZE;
	  else curframes = frames-i;
	  mus_sound_read(fd, 0, curframes-1, chans, bufs); 
	  if (chans == 1)
	    {
	      for (k=0;k<curframes;k++) obuf[k] = MUS_SAMPLE_TO_SHORT(bufs[0][k]);
	    }
	  else
	    {
	      if (chans == 2)
		{
		  for (k=0, n=0;k<curframes;k++, n+=2) 
		    {
		      obuf[n] = MUS_SAMPLE_TO_SHORT(bufs[0][k]); 
		      obuf[n+1] = MUS_SAMPLE_TO_SHORT(bufs[1][k]);
		    }
		}
	      else
		{
		  for (k=0, j=0;k<curframes;k++, j+=chans)
		    {
		      for (n=0;n<chans;n++) obuf[j+n] = MUS_SAMPLE_TO_SHORT(bufs[n][k]);
		    }
		}
	    }
	  if (afd == -1)
	    {
#if defined(LINUX) && defined(PPC)
	      afd = mus_audio_open_output(output_device, srate, chans, MUS_COMPATIBLE_FORMAT, 0);
#else
	      afd = mus_audio_open_output(output_device, srate, chans, MUS_COMPATIBLE_FORMAT, outbytes);
#endif
	      if (afd == -1) return(-1);
	    }
	  outbytes = curframes * chans * 2;
	  mus_audio_write(afd, (char *)obuf, outbytes);
	}
      if (afd != -1) mus_audio_close(afd);
      mus_sound_close_input(fd);
      for (i=0; i<chans; i++) FREE(bufs[i]);
      FREE(bufs);
      FREE(obuf);
    }
  else return(-2);
  return(0);
}

#ifndef MUS_INTERP_LINEAR
  #define MUS_INTERP_LINEAR 0
#endif

void fill_locsig(double *arr, int chans, double degree, double scaler, int type)
{
  mus_fill_locsig(arr, chans, degree, scaler, type);
}

/* type conversions for lisp FFI's that can't handle off_t */

/* ffi.lisp */
int clm_sound_samples(const char *arg) {return((int)mus_sound_samples(arg));}
int clm_sound_frames(const char *arg) {return((int)mus_sound_frames(arg));}
int clm_sound_data_location(const char *arg) {return((int)mus_sound_data_location(arg));}
int clm_sound_length(const char *arg) {return((int)mus_sound_length(arg));}

int clm_sound_reopen_output(const char *arg, int chans, int format, int type, int data_loc)
{
  return(mus_sound_reopen_output(arg, chans, format, type, (off_t)data_loc));
}
int clm_sound_close_output(int fd, int bytes_of_data) {return(mus_sound_close_output(fd, (off_t)bytes_of_data));}
int clm_sound_seek_frame(int tfd, int frame) {return((int)mus_sound_seek_frame(tfd, (off_t)frame));}
int clm_file_write_zeros(int tfd, int num) {return((int)mus_file_write_zeros(tfd, (off_t)num));}

/* sndlib2clm.lisp */
int clm_sound_comment_start(const char *arg) {return((int)mus_sound_comment_start(arg));}
int clm_sound_comment_end(const char *arg) {return((int)mus_sound_comment_end(arg));}
int clm_sound_override_header(const char *arg, int srate, int chans, int format, int type, int location, int size)
{
  return((int)mus_sound_override_header(arg, srate, chans, format, type, (off_t)location, (off_t)size));
}
int clm_file_seek_frame(int tfd, int frame) {return((int)mus_file_seek_frame(tfd, (off_t)frame));}
int clm_header_samples(void) {return((int)mus_header_samples());}
int clm_header_data_location(void) {return((int)mus_header_data_location());}
int clm_header_comment_start(void) {return((int)mus_header_comment_start());}
int clm_header_comment_end(void) {return((int)mus_header_comment_end());}
int clm_header_true_length(void) {return((int)mus_header_true_length());}
int clm_samples_to_bytes(int format, int size) {return(mus_samples_to_bytes(format, (off_t)size));}
int clm_bytes_to_samples(int format, int size) {return(mus_bytes_to_samples(format, (off_t)size));}
int clm_header_write(const char *name, int type, int srate, int chans, int loc, int size_in_samples, int format, const char *comment, int len)
{
  return((int)mus_header_write(name, type, srate, chans, (off_t)loc, (off_t)size_in_samples, format, comment, len));
}
int clm_header_aux_comment_start(int n) {return((int)mus_header_aux_comment_start(n));}
int clm_header_aux_comment_end(int n) {return((int)mus_header_aux_comment_end(n));}

int clm_sound_maxamp(const char *ifile, int chans, double *vals, int *times)
{
  off_t *ts;
  mus_sample_t *xs;
  int i;
  int res = -1;
  ts = (off_t *)calloc(chans, sizeof(off_t));
  xs = (mus_sample_t *)calloc(chans, sizeof(mus_sample_t));
  res = mus_sound_maxamps(ifile, chans, xs, ts);
  for (i = 0; i < chans; i++) {times[i] = (int)ts[i]; vals[i] = (double)xs[i];}
  free(ts);
  free(xs);
  return(res);
}


/* endian converters (clm1.lisp) */
void swap_int_array(int *arr, int size)
{
  int i, lim;
  unsigned char *p;
  unsigned char temp;
  p = (unsigned char *)arr;
  lim = size * 4;
  for (i = 0; i < lim; i += 4)
    {
      temp = p[i + 0];
      p[i + 0] = p[i + 3]; 
      p[i + 3] = temp;
      temp = p[i + 1];
      p[i + 1] = p[i + 2]; 
      p[i + 2] = temp;
    }
}

void swap_double_array(double *arr, int size)
{
  int i, lim;
  unsigned char *p;
  unsigned char temp;
  p = (unsigned char *)arr;
  lim = size * 8;
  for (i = 0; i < lim; i += 8)
    {
      temp = p[i + 0];
      p[i + 0] = p[i + 7]; 
      p[i + 7] = temp;
      temp = p[i + 1];
      p[i + 1] = p[i + 6]; 
      p[i + 6] = temp;
      temp = p[i + 2];
      p[i + 2] = p[i + 5]; 
      p[i + 5] = temp;
      temp = p[i + 3];
      p[i + 3] = p[i + 4];
      p[i + 4] = temp;
    }
}
