/* rcp.c
 * run clm note list (with-psound) as a separate C program.
 */

#include <config.h>
#include <math.h>
#include <stdio.h>
#if HAVE_FCNTL_H
  #include <fcntl.h>
#endif
#if HAVE_LIMITS_H
  #include <limits.h>
#endif
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#if defined(SGI) || defined(LINUX) || defined(SUN) || defined(ALPHA)
  #include <sys/time.h>
#endif

#if (defined(HAVE_LIBC_H) && (!defined(HAVE_UNISTD_H)))
  #include <libc.h>
#else
  #ifndef _MSC_VER
    #include <unistd.h>
  #endif
#endif

#ifdef SGI
#include <sys/fpu.h>
#endif

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#ifndef CLM_FILE_BUFFER_SIZE
  #define CLM_FILE_BUFFER_SIZE 65536
#endif

#define CLM_SHM_KEY 0x00636c6d
#define FASMIX -2
#define REVERB -1
#define CLM_INTERRUPTED 0
#define CURRENT_OUTPUT_FILE 2
#define REVERB_FILE 3

static const char I_dcdt[4] = {'.','c','d','t'};

typedef int (*intinsfun)();
static intinsfun *ins_pool;        /* entry points of instruments */

#include "rc.pool"

static int match_four_chars(unsigned char *head, const unsigned char *match)
{ 
  int i;
  for (i=0;i<4;i++) if (head[i] != match[i]) return(0);
  return(1);
}
  
static int as_int(int fd, unsigned char *buf)
{
  read(fd,buf,4);
  return(mus_char_to_bint(buf));
}

static float as_float(int fd, unsigned char *buf)
{
  read(fd,buf,4);
  return(mus_char_to_bfloat(buf));
}

static int last_string_len;

static char *as_string(int fd, unsigned char *buf)
{
  int len;
  char *str;
  len=as_int(fd,buf);
  last_string_len=len;
  str = (char *)calloc(len+1,sizeof(char));
  read(fd,str,len);
  return(str);
}

static int last_int_len;

static int *as_int_array(int fd, unsigned char *buf, int *data, int *data_len)
{
  int len;
#if MUS_LITTLE_ENDIAN
  int i;
  unsigned char *dp;
#endif
  len = as_int(fd,buf);
  last_int_len = len;
  if (len > (*data_len))
    {
      if ((*data_len) != 0) free(data);
      (*data_len) = (2*len);
      data = (int *)calloc((*data_len),sizeof(int));
    }
  read(fd,(char *)data,len*4);
#if MUS_LITTLE_ENDIAN
  for (i=0,dp=(unsigned char *)data;i<len;i++,dp+=4) data[i]=mus_char_to_bint(dp);
#endif
  return(data);
}

static float *as_float_array(int fd, unsigned char *buf, float *data, int *data_len)
{
  int len;
#if MUS_LITTLE_ENDIAN
  int i;
  unsigned char *dp;
#endif
  len = as_int(fd,buf);
  if (len > (*data_len))
    {
      if ((*data_len) != 0) free(data);
      (*data_len) = (2*len);
      data = (float *)calloc((*data_len),sizeof(float));
    }
  read(fd,(char *)data,len*4);
#if MUS_LITTLE_ENDIAN
  for (i=0,dp=(unsigned char *)data;i<len;i++,dp+=4) data[i]=mus_char_to_bfloat(dp);
#endif
  return(data);
}

static void update_io(int *outb, int *datain, int adr)
{
  outb[CLM_IO_DATA_END] = datain[CLM_IO_DATA_END+adr];
  outb[CLM_IO_DATA_START] = datain[CLM_IO_DATA_START+adr];
  outb[CLM_IO_SIZE] = datain[CLM_IO_SIZE+adr];
  outb[CLM_IO_BEG] = datain[CLM_IO_BEG+adr];
  outb[CLM_IO_END] = datain[CLM_IO_END+adr];
}

static int *initialize_io(int outf_fd, int ochans, int ohdrend)
{
  int *outf;
  int i;
  outf = (int *)calloc(CLM_IO_DATS+ochans+6,sizeof(int));
  outf[0] = 37;
  outf[CLM_IO_FD] = outf_fd;                        /* set up CLM's view of the output file */
  outf[CLM_IO_CHANS] = ochans;
  outf[CLM_IO_SIZE]=0;
  outf[CLM_IO_BEG]=0;
  outf[CLM_IO_END]=1023;
  outf[CLM_IO_BUFSIZ]=1024;
  outf[CLM_IO_DATA_START]=0;
  outf[CLM_IO_DATA_END]=0;
  outf[CLM_IO_DIR]=CLM_IO_OUT_FILE;
  outf[CLM_IO_HDR_END]=ohdrend;
  outf[CLM_IO_DATS+CLM_AREF_IBLOCK]=CLM_IO_DATS+6;
  outf[CLM_IO_DATS+CLM_AREF_TYPE]=39;
  outf[CLM_IO_DATS+CLM_AREF_SIZE]=ochans;
  outf[CLM_IO_DATS+CLM_AREF_DIMS]=1;
  for (i=0;i<ochans;i++) outf[CLM_IO_DATS+6+i]=(int)MUS_MAKE_SAMPLE_ARRAY(1024);
  return(outf);
}

static int open_files[32];
static int open_file_ctr = 0;
int rc_open(char *ifile, int *datai, int loc);

int rc_open(char *ifile, int *datai, int loc)
{
  int ifd,ichans,i;
  if ((ifd=mus_file_open_read(ifile)) == -1) return(-1);
  ichans = mus_sound_chans(ifile);
  mus_file_open_descriptors(ifd,ifile,
			    mus_sound_data_format(ifile),mus_sound_datum_size(ifile),mus_sound_data_location(ifile),
			    ichans,mus_sound_header_type(ifile));
  datai[loc+CLM_IO_DATS+CLM_AREF_IBLOCK]=loc+CLM_IO_DATS+6;
  datai[loc+CLM_IO_DATS+CLM_AREF_TYPE]=39;
  datai[loc+CLM_IO_DATS+CLM_AREF_SIZE]=ichans;
  datai[loc+CLM_IO_DATS+CLM_AREF_DIMS]=1;
  datai[loc+CLM_IO_END] = -1;
  for (i=0;i<ichans;i++) 
    {
      datai[loc+CLM_IO_DATS+6+i] = (int)MUS_MAKE_SAMPLE_ARRAY(CLM_FILE_BUFFER_SIZE);
    }
  open_files[open_file_ctr++] = ifd;
  return(ifd);
}


#define CDT_BUFFER_SIZE 65536

int main (int argc, char *argv[])
{
  /* read .cdt file, producing CLM output sound.
   *
   * argv[1] = <mumble.cdt> (our current data file)
   *
   * ----------------
   * ".cdt"
   * srate
   * scaled-to value
   * output file name length (0 if direct to dac?)
   *   output file name
   *   output channels
   *   output data format
   *   output type
   *   output header end
   * output header length
   *   output header
   * reverb file name length (0 if none)
   *   reverb file name
   *   reverb channels
   *   reverb data format
   *   reverb type
   *   reverb header end
   * reverb header length
   *   reverb header
   * stats flag
   * final data format (for scaled-to)
   * blocks describing instrument calls, each call:
   *   index into instrument function array (ins_pool in rc.pool)
   *   beg
   *   end
   *   datar length (floats)
   *   datar
   *   datai length (ints)
   *   datai
   * EOF or -1 = signal that next block is reverb call
   *   possible reverb instrument call (just like above)
   * ----------------
   *
   * each with-sound call, if rc is in effect, opens the .cdt file, sets up the header info,
   *   then on each instrument call, we write the data to the .cdt array rather than call the function.
   *   When all the data is out (at end of with-sound), we call rc with the .cdt name.
   * We also create upon with-sound starting up a list of currently active instruments from
   *   *clm-instruments* which has the associated C file and function names.  This file is
   *   written/read alongside the .cdt file.
   */

  int *outf,*revf;
  int *datai;
  float *datar;
  double *ins_datar;
  int idl;
  float scaled_to;
  char *cdt_filename,*outf_filename,*revf_filename,*final_comment;
  unsigned char *cdt_buffer;
  int cdt_fd,revf_fd,ochans,rchans,i,j,odataformat,rdataformat,otype,rtype,len,srate,stats,bytes,ohdrend,rhdrend;
  int beg,end,p_index,happy,cur_datar_len,cur_datai_len,cur_matrix_len,outf_adr,revf_adr,inf_len;
  int final_format,last_datai_len;
  int shm_size;

#ifdef SGI
  union fpc_csr f;
  f.fc_word = get_fpc_csr();
  f.fc_struct.flush = 1;
  set_fpc_csr(f.fc_word);
#endif

  cur_datar_len = 0;
  cur_datai_len = 0;
  cur_matrix_len = 0;
  revf_fd = 0;


  mus_sound_initialize();
  initialize_cmus();
  ins_pool_initialize();

  cdt_filename=argv[1]; 
  cdt_fd = open(cdt_filename,O_RDONLY);
  if (cdt_fd == -1) {printf("can't open clm no-ffi data file: %s",cdt_filename); exit(1);}
  cdt_buffer = (unsigned char *)calloc(CDT_BUFFER_SIZE,sizeof(unsigned char));
  read(cdt_fd,cdt_buffer,4);
  if (!(match_four_chars((char *)cdt_buffer,I_dcdt))) {printf("%s does not start with .cdt!",cdt_filename); exit(1);}

  /* lots of useless stuff here, but we need to read it to get to the instrument calls */

  srate = as_int(cdt_fd,cdt_buffer);
  scaled_to = as_float(cdt_fd,cdt_buffer);
  outf_filename = as_string(cdt_fd,cdt_buffer);
  ochans = as_int(cdt_fd,cdt_buffer);
  odataformat = as_int(cdt_fd,cdt_buffer);
  odataformat = (odataformat & 0xffff);
  otype = as_int(cdt_fd,cdt_buffer);
  ohdrend = as_int(cdt_fd,cdt_buffer);
  final_comment = as_string(cdt_fd,cdt_buffer);
  mus_file_open_descriptors(MUS_DAC_CHANNEL, "dac",odataformat, 2, 0, ochans, 0);
  outf = initialize_io(MUS_DAC_CHANNEL,ochans,0);
  mus_set_rt_audio_p(open_dac(srate,ochans,4096,0));

  /* (if (> rec-chans 0) (set-rt-record-p (rt-open-adc srate rec-chans line))) */

  len = as_int(cdt_fd,cdt_buffer);              /* reverb file name length (0=no reverb) */
  if (len > 0)
    {
      revf_filename = (char *)calloc(len,sizeof(char));
      read(cdt_fd,revf_filename,len);   
      rchans = as_int(cdt_fd,cdt_buffer);
      rdataformat = as_int(cdt_fd,cdt_buffer);
      rtype = as_int(cdt_fd,cdt_buffer);
      rhdrend = as_int(cdt_fd,cdt_buffer);
      len = as_int(cdt_fd,cdt_buffer);
      read(cdt_fd,cdt_buffer,len);
      mus_file_open_descriptors(MUS_DAC_REVERB,"reverb",odataformat,2,0,rchans,0);
    }
  stats = as_int(cdt_fd,cdt_buffer);
  final_format = as_int(cdt_fd,cdt_buffer);

  /* final_format useless (I hope), so it's now actually shm mem size (if needed) */
  shm_size = final_format;
  if (shm_size) clm_attach_controls(shm_size);
  happy = 1;
  while (happy)
    {
      bytes = read(cdt_fd,cdt_buffer,4);        /* p_index */
      if (bytes < 4) happy = 0;                 /* EOF = all done */
      else
	{
	  p_index = mus_char_to_bint(cdt_buffer);
	  beg = as_int(cdt_fd,cdt_buffer);
	  end = as_int(cdt_fd,cdt_buffer);
	  datar = as_float_array(cdt_fd,cdt_buffer,datar,&cur_datar_len);
	  ins_datar = (double *)calloc(cur_datar_len,sizeof(double));
	  for (idl=0;idl<cur_datar_len;idl++) ins_datar[idl] = (double)(datar[idl]);

	  datai = as_int_array(cdt_fd,cdt_buffer,datai,&cur_datai_len);
	  last_datai_len = last_int_len;

	  outf_adr = datai[CURRENT_OUTPUT_FILE];
	  if (outf_adr != 0)
	    {
	      for (i=outf_adr,j=0;i<(outf_adr+CLM_IO_DATS+6+ochans);i++,j++) datai[i]=outf[j];
	      datai[outf_adr+CLM_IO_DATS+CLM_AREF_IBLOCK] = outf_adr+CLM_IO_DATS+6;
	    }
	  (*ins_pool[p_index])(beg,end,ins_datar,cur_datar_len,datai,cur_datai_len); 
	  free(ins_datar);
	  update_io(outf,datai,outf_adr);
	}
    }
  close(cdt_fd);
  /* now run all ins to completion */
  clm_finalize_pool(outf_adr,FALSE,datai);
  close_dac();
  close_adc();
  if (shm_size) clm_detach_controls();
  return(0);
}
